//******************************************************************************
//
// Copyright (c) 2008 Advantech Industrial Automation Group.
//
// advantech watch dog driver for WinBond W83627HF chip
// 
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation; either version 2 of the License, or (at your option) 
// any later version.
// 
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 
// Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
//
//
//******************************************************************************
/*
 *
 * File:        advwdt627.c
 * Version:     1.11.1 <05/10/2009>
 * Author:      WangMao
 *
 * Description: The watchdog driver is for WinBond W83627HF chip to operate 
 *           Advantech UNO-2050E,UNO2052E,UNO-2170 etc. 
 *           The structure of this driver comes mainly from kernel source:
 *           drivers/char/advantechwdt.c
 *
 * Status:      working
 *
 * Change Log: 
 *            Version 1.00 <03/08/2006> WangMao        
 *                 - Initial version  
 *            Version 1.01 <07/10/2006> WangMao
 *                 - Add better support for kernel parameter on both 2.4 and 2.6 kernel.
 *            Version 1.10 <09/20/2006> WangMao
 *                 - Add IO address detection (0x2E or 0x4E?)
 *            Version 1.11 <04/01/2009> jianfeng
 *                 - to support FC8,FC9,openSuse etc.
 *            Version 1.12 <05/06/2010> jianfeng
 *      	     to support FC10,FC11,FC12,SUSE 11.2 Madriva 2010
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  This software is provided "as is" without express or
 * implied warranty.
 * ===========================================================================*/
#include <linux/version.h>
#ifndef KERNEL_VERSION
#define  KERNEL_VERSION(a, b, c) KERNEL_VERSION((a)*65536+(b)*256+(c))
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/version.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/fs.h>

static unsigned long advwdt_is_open;
static unsigned long timeout = 0;
static char adv_expect_close;
static int  advwdt_base_addr = 0x2E;
static int default_period = 45;

#define ADVANTECH_WDT_VER        "1.12"
#define ADVANTECH_WDT_DATE       "05/06/2010"

#define WATCHDOG_MINOR	130

//#define ADVANTECH_DEBUG
#ifdef ADVANTECH_DEBUG
#define DEBUGPRINT     printk
#else
#define DEBUGPRINT(a, ...)
#endif

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif
//WangMao add the below line for uno2053e is special: it uses the system timer
//as its watchdog timer,so there is no need to request and release the io space of watchdog
static int is_uno2053e = 0;

//WangMao add this macro to support module parameter on both 2.6 and 2.4 kernel better
//@06/07/10
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(nowayout,"i");
#else
module_param(nowayout, int, 0);
#endif
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once start(default=CONFIG_WATCHDOG_NOWAYOUT)");

#ifndef MODULE
static int __init adv_setup(char *str)
{
    int ints[4];
    str = get_options(str, ARRAY_SIZE(ints), ints);

    if(ints[0] > 0)
    {
        advwdt_base_addr = ints[1];
            if(ints[0] > 1)
            advwdt_base_addr = ints[2];
    }
    return 1;
}

__setup("advwdt=", adv_setup);

#endif /* !MODULE */
static void advwdt_wdtaccess ( int offset )
{
    unsigned short usLed;
    unsigned short usData;

    // Unlock W83627HF
    outb( 0x87, advwdt_base_addr );
    outb( 0x87, advwdt_base_addr );        	
    udelay( 10 );
        
    // Select registers of watchdog timer
    outb( 0x07, advwdt_base_addr );
    outb(0x08 , advwdt_base_addr + 1 );
    udelay( 10 );

    // Enable the functions of watchdog timer
    outb( 0x30, advwdt_base_addr );
    udelay( 5 );
    outb( 0x01, advwdt_base_addr + 1 );        
    udelay( 10 );

    // If have to enable PLED flashing
    if(offset & 0xff)
        usLed = 0x80; // 1Hz Toggle Pulse
    else
        usLed = 0x40; // Disable PLED flashing

    // Set second as counting unit
    outb( 0xf5, advwdt_base_addr );
    usData = inb( advwdt_base_addr + 1 ) & ~0xc0; // remove PLED setting
    usData |= usLed; // set PLED flash if any
    outb( usData & ~0x08, advwdt_base_addr + 1 );        
    udelay( 10 );

    // Set timeout interval and start counting if no-zero 'Offset'
    outb( 0xf6, advwdt_base_addr);
    outb( offset & 0xff ,advwdt_base_addr + 1 );

    // Lock W83627HF
    outb( 0xaa, advwdt_base_addr );
}

static ssize_t
advwdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    /*  Can't seek (pwrite) on this device  */
    #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,8)
        if (ppos != &file->f_pos)
            return -ESPIPE;
    #endif

    if (count) 
    {
        if (!nowayout) 
        {
            size_t i;
            adv_expect_close = 0;
            for (i = 0; i != count; i++) 
            {
                char c;
                if(get_user(c, buf+i))
                    return -EFAULT;
                if (c == 'V')
                    adv_expect_close = 42;
            }
        }

    advwdt_wdtaccess( timeout );
    }
    return count;
}

static int
advwdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	  unsigned long arg)
{
    int retval = 0;
    unsigned long new_timeout;
    static struct watchdog_info ident = 
    {
        #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
            options:            WDIOF_KEEPALIVEPING | WDIOF_SETTIMEOUT,    /* | WDIOF_MAGICCLOSE,*/ 
        #else
            options:            WDIOF_KEEPALIVEPING | WDIOF_SETTIMEOUT | WDIOF_MAGICCLOSE, 

        #endif
            firmware_version:     0,
            identity:            "Advantech WDT"
    };

    switch ( cmd ) 
    {
    case WDIOC_GETSUPPORT:
        if (copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident)))
            return -EFAULT;
        break;

    case WDIOC_GETSTATUS:
    case WDIOC_GETBOOTSTATUS:
        return put_user(0, (int *)arg);

    case WDIOC_KEEPALIVE:
        advwdt_wdtaccess( timeout );
        break;

    case WDIOC_SETTIMEOUT:
        if (get_user( new_timeout, (unsigned long *)arg))
            return -EFAULT;
        if( (new_timeout>255) || (new_timeout<0) )   
        {
            printk(KERN_CRIT "Advantch WDT: the input timeout is out of range.\n");
            printk(KERN_CRIT "Please choose valid one between 0 ~ 255 seconds.\n");
            return -EINVAL; 
        }
        else if ( retval == 1 )	
        { 
            advwdt_wdtaccess( 0 );
            timeout = 0;
            return -EINVAL;	             
        }
        else
        {
            timeout = new_timeout;
            advwdt_wdtaccess( new_timeout );
        }
        break;
        
        case WDIOC_GETTIMEOUT: 
        if( timeout==0 )   	   
        {
            return -EFAULT;
        }         
        return put_user( timeout, (unsigned long *)arg);
        
        case WDIOC_SETOPTIONS:	
        {
            int options, retval = -EINVAL;
            if (get_user(options, (int *)arg))
            return -EFAULT;

            if (options & WDIOS_DISABLECARD) 
            {
                if (!nowayout)
	   {
                    advwdt_wdtaccess( 0 );
                    printk("WDT disabled.\n");
                }
                else
                {
                    printk("CONFIG_WATCHDOG_NOWAYOUT is enabled. WDT can't be disabled.\n");
                    return -EFAULT;
                }
                retval = 0;
            }

            if (options & WDIOS_ENABLECARD) 
            {           
                if ( timeout== 0 )
                timeout = default_period;    //set 45 as default value
                advwdt_wdtaccess( timeout );
                printk("Enable the watchdog!\n The timeout is: %ld\n", timeout );
                DEBUGPRINT( "Enable the watchdog!\n The timeout index is: %d\n", timeout );
                retval = 0;
            }
            return retval;      
        }
        break;

        default:
            return -ENOTTY;
        }
        return 0;
}

static int
advwdt_open(struct inode *inode, struct file *file)
{
    if (test_and_set_bit(0, &advwdt_is_open))
        return -EBUSY;
    advwdt_wdtaccess( timeout );
    return 0;
}

static int
advwdt_close(struct inode *inode, struct file *file)
{
    if (adv_expect_close == 42) 
    {
        advwdt_wdtaccess( 0 );
    } 
    else 
    {
        printk(KERN_CRIT "Advantech WDT: Unexpected close, not stopping watchdog!\n");
        advwdt_wdtaccess( timeout );
    }
    clear_bit(0, &advwdt_is_open);
    adv_expect_close = 0;
    return 0;
}

/*
 *	Notifier for system down
 */
static int
advwdt_notify_sys(struct notifier_block *this, unsigned long code, void *unused)
{
    if (code == SYS_DOWN || code == SYS_HALT) {
    /* Turn the WDT off */
    printk("sys shutdown be notified by advwdt_notify_sys() \n");
    advwdt_wdtaccess( 0 );
}
    return NOTIFY_DONE;
}
 
/*
 *	Kernel Interfaces
 */
static struct file_operations advwdt_fops = {
    owner:		THIS_MODULE,
	  write:		advwdt_write,
	  ioctl:		advwdt_ioctl,
	  open:		  advwdt_open,
	  release:	advwdt_close,
};
 
static struct miscdevice advwdt_miscdev = {
	  minor:		WATCHDOG_MINOR,
    name:		  "watchdog",			
	  fops:		  &advwdt_fops,
};

/*
 *	The WDT needs to learn about soft shutdowns in order to
 *	turn the timebomb registers off.
 */ 
static struct notifier_block advwdt_notifier = {
	advwdt_notify_sys,
	NULL,
	0
};

static int __init advwdt_init(void)
{        
    int type;
    
    misc_register(&advwdt_miscdev);
    request_region(advwdt_base_addr, 2, "W83627HF");
    
    //WangMao add the below to proceed a io address detection @06/09/20
    //original situation: io address is 0x2E    
    // Unlock W83627HF
    outb( 0x87, advwdt_base_addr );
    outb( 0x87, advwdt_base_addr );        	
    udelay( 10 );
        
    // Select registers of watchdog timer
    outb( 0x07, advwdt_base_addr );
    outb( 0x08 , advwdt_base_addr + 1 );
    udelay( 10 );
    
    //DEV ID
    outb( 0x20 , advwdt_base_addr );
    type = inb( advwdt_base_addr+1 );
    //Lock
    outb( 0xAA,advwdt_base_addr);
    
    //new situation: io addresss: 0x4E
    if( type != 0x52 && type != 0xA0 )
    {
    	advwdt_base_addr = 0x4E;
	//if is uno2053e,the request of 4e will fail for the system timer uses it
    	if( request_region( advwdt_base_addr,2,"W83627HF" ) == NULL )
		is_uno2053e = 1;
    	
    	// Unlock W83627HF
    	outb( 0x87, advwdt_base_addr );
    	outb( 0x87, advwdt_base_addr );        	
    	udelay( 10 );
        
    	// Select registers of watchdog timer
    	outb( 0x07, advwdt_base_addr );
    	outb( 0x08 , advwdt_base_addr + 1 );
    	udelay( 10 );
    	
	//DEV ID
    	outb( 0x20 , advwdt_base_addr );
    	type = inb( advwdt_base_addr+1 );
    	//Lock
    	outb( 0xAA,advwdt_base_addr);
    	if( type != 0x52 )	//not found W83627HF
    	{
    		printk("not found W83627HF!\n");
    		release_region( advwdt_base_addr, 2 );
    		return -ENODEV;
    	}
    }
    //WangMao add end
    
    register_reboot_notifier(&advwdt_notifier);
    printk( "========================================================\n");
    printk( "         WDT driver for Winbond 83627HF Chip \n");
    printk( "                   V%s [%s]\n", ADVANTECH_WDT_VER, ADVANTECH_WDT_DATE);
    printk( "            Advantch eAutomation Devision.\n");
    printk( "========================================================\n");
    printk( "CONFIG_WATCHDOG_NOWAYOUT is currently %s\n", nowayout?"enabled!":"disabled!");
    //WangMao add to display IO address @06/09/20
    printk( "The IO address of 83627HF is: 0x%x\n",advwdt_base_addr);
    return 0;
}

static void __exit advwdt_exit(void)
{
    DEBUGPRINT("--->>>advwdt_exit!\n");
    //uno2053e: the iospace is owned by system timer,no need to release it
    if( !is_uno2053e )
        release_region(advwdt_base_addr, 2);
    misc_deregister(&advwdt_miscdev);
    unregister_reboot_notifier(&advwdt_notifier);
}

module_init(advwdt_init);
module_exit(advwdt_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("WangMao");
MODULE_DESCRIPTION("Advantech WDT driver");



