//******************************************************************************
//
// Copyright (c) 2009 Advantech Industrial Automation Group.
//
// advantech watch dog driver for SCH311X chip
// 
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation; either version 2 of the License, or (at your option) 
// any later version.
// 
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 
// Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
//
//
//******************************************************************************
/*
 *
 * File:        advwdt311x.c
 * Author:      james dai
 *
 * Description: The watchdog driver is for SCH311X chip to operate 
 *              Advantech UNO-2173,TPC-1770 etc. 
 *              The structure of this driver comes mainly from kernel source:
 *              drivers/char/advantechwdt.c
 *
 * Status:      working
 *
 * Change Log: 
 *              Version 1.00 <08/19/2008> james dai        
 *              - Initial version  
 *              Version 1.01 <05/06/2010> jianfeng
 *      	     to support FC10,FC11,FC12,SUSE 11.2 Madriva 2010

 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  This software is provided "as is" without express or
 * implied warranty.
 * ===========================================================================*/
#include <linux/version.h>
#ifndef KERNEL_VERSION
#define  KERNEL_VERSION(a, b, c) KERNEL_VERSION((a)*65536+(b)*256+(c))
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/version.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/fs.h>

static unsigned long advwdt_is_open;
static unsigned long timeout = 0;
static char adv_expect_close;
static int  advwdt_base_addr = 0x800;
static int  advwdt_base_len = 0x70;
static int default_period = 45;

#define ADVANTECH_WDT_VER        "1.01"
#define ADVANTECH_WDT_DATE       "05/06/2010"

#define WATCHDOG_MINOR	130

//#define ADVANTECH_DEBUG
#ifdef ADVANTECH_DEBUG
#define DEBUGPRINT     printk
#else
#define DEBUGPRINT(a, ...)
#endif

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(nowayout,"i");
#else
module_param(nowayout, int, 0);
#endif
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once start(default=CONFIG_WATCHDOG_NOWAYOUT)");



static void advwdt_wdtaccess ( int timeout )
{

	   // program GPIO_52 as WDT;
	   outb( 0x0C,advwdt_base_addr + 0x47);

	   // WDT Time-out value units select //WDT_TIMEOUT = 0x65;
	   // Set second as counting unit
	   outb(0x80,advwdt_base_addr + 0x65);

	   // WDT Time-out value  //WDT_VAL = 0x66; 0: disable 1~255: n second timeout
	   outb(  timeout & 0xff,advwdt_base_addr + 0x66);

	   // Enable WDT function  //WDT_CTRL = 0x68;
	  outb(  0x00,advwdt_base_addr + 0x68);
}

static ssize_t
advwdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
   /*  Can't seek (pwrite) on this device  */
   #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,8)
   if (ppos != &file->f_pos)
      return -ESPIPE;
   #endif

   if (count) 
   {
      if (!nowayout) 
      {
         size_t i;
         adv_expect_close = 0;
         for (i = 0; i != count; i++) 
         {
            char c;
            if(get_user(c, buf+i))
                    return -EFAULT;
            if (c == 'V')
            adv_expect_close = 42;
         }
      }
      advwdt_wdtaccess( timeout );
   }
   return count;
}

static int
advwdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	  unsigned long arg)
{
   int retval = 0;
   unsigned long new_timeout;
   static struct watchdog_info ident = 
   {
      #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
         options:            WDIOF_KEEPALIVEPING | WDIOF_SETTIMEOUT,    /* | WDIOF_MAGICCLOSE,*/ 
      #else
         options:            WDIOF_KEEPALIVEPING | WDIOF_SETTIMEOUT | WDIOF_MAGICCLOSE, 
      #endif
         firmware_version:     0,
      identity:            "Advantech WDT"
   };

   switch ( cmd ) 
   {
      case WDIOC_GETSUPPORT:
         if (copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident)))
            return -EFAULT;
         break;

      case WDIOC_GETSTATUS:
      case WDIOC_GETBOOTSTATUS:
        return put_user(0, (int *)arg);

      case WDIOC_KEEPALIVE:
        advwdt_wdtaccess( timeout );
        break;

      case WDIOC_SETTIMEOUT:
         if (get_user( new_timeout, (unsigned long *)arg))
             return -EFAULT;
         if( (new_timeout>255) || (new_timeout<0) )   
         {
            printk(KERN_CRIT "Advantch WDT: the input timeout is out of range.\n");
            printk(KERN_CRIT "Please choose valid one between 0 ~ 255 seconds.\n");
            return -EINVAL; 
         }
         else if ( retval == 1 )	
         { 
            advwdt_wdtaccess( 0 );
            timeout = 0;
            return -EINVAL;	             
         }
         else
         {
            timeout = new_timeout;
            advwdt_wdtaccess( new_timeout );
         }
         break;
        
      case WDIOC_GETTIMEOUT: 
         if( timeout==0 )   	   
         {
            return -EFAULT;
         }         
         return put_user( timeout, (unsigned long *)arg);
        
      case WDIOC_SETOPTIONS:	
      {
         int options, retval = -EINVAL;
         if (get_user(options, (int *)arg))
            return -EFAULT;

         if (options & WDIOS_DISABLECARD) 
         {
            if (!nowayout)
            {
               advwdt_wdtaccess( 0 );
               printk("WDT disabled.\n");
            }
            else
            {
               printk("CONFIG_WATCHDOG_NOWAYOUT is enabled. WDT can't be disabled.\n");
               return -EFAULT;
            }
            retval = 0;
         }

         if (options & WDIOS_ENABLECARD) 
         {           
            if ( timeout== 0 )
               timeout = default_period;    //set 45 as default value
            advwdt_wdtaccess( timeout );
            printk("Enable the watchdog!\n The timeout is: %ld\n", timeout );
            DEBUGPRINT( "Enable the watchdog!\n The timeout index is: %d\n", timeout );
            retval = 0;
         }
         return retval;      
      }
      default:
         return -ENOTTY;
   }
   return 0;
}

static int
advwdt_open(struct inode *inode, struct file *file)
{
    if (test_and_set_bit(0, &advwdt_is_open))
        return -EBUSY;
    advwdt_wdtaccess( timeout );
    return 0;
}

static int
advwdt_close(struct inode *inode, struct file *file)
{
    if (adv_expect_close == 42) 
    {
        advwdt_wdtaccess( 0 );
    } 
    else 
    {
        printk(KERN_CRIT "Advantech WDT: Unexpected close, not stopping watchdog!\n");
        advwdt_wdtaccess( timeout );
    }
    clear_bit(0, &advwdt_is_open);
    adv_expect_close = 0;
    return 0;
}

/*
 *	Notifier for system down
 */
static int
advwdt_notify_sys(struct notifier_block *this, unsigned long code, void *unused)
{
    if (code == SYS_DOWN || code == SYS_HALT) {
    /* Turn the WDT off */
    printk("sys shutdown be notified by advwdt_notify_sys() \n");
    advwdt_wdtaccess( 0 );
}
    return NOTIFY_DONE;
}
 
/*
 *	Kernel Interfaces
 */
static struct file_operations advwdt_fops = {
   owner:      THIS_MODULE,
   write:      advwdt_write,
   ioctl:      advwdt_ioctl,
   open:       advwdt_open,
   release:    advwdt_close,
};
 
static struct miscdevice advwdt_miscdev = {
   minor:   WATCHDOG_MINOR,
   name:    "watchdog",			
   fops:    &advwdt_fops,
};

/*
 *	The WDT needs to learn about soft shutdowns in order to
 *	turn the timebomb registers off.
 */ 
static struct notifier_block advwdt_notifier = {
	advwdt_notify_sys,
	NULL,
	0
};

static int __init advwdt_init(void)
{        
    misc_register(&advwdt_miscdev);
    request_region(advwdt_base_addr,advwdt_base_len, "SCH311X");
    
    register_reboot_notifier(&advwdt_notifier);
    printk( "========================================================\n");
    printk( "         WDT driver for SCH311X Chip \n");
    printk( "    V%s [%s]\n", ADVANTECH_WDT_VER, ADVANTECH_WDT_DATE);
    printk( "         Advantch eAutomation Devision.\n");
    printk( "========================================================\n");
    printk( "CONFIG_WATCHDOG_NOWAYOUT is currently %s\n", nowayout?"enabled!":"disabled!");

    printk( "The IO address of SCH311X is: 0x%x\n",advwdt_base_addr);
    return 0;
}

static void __exit advwdt_exit(void)
{
    DEBUGPRINT("--->>>advwdt_exit!\n");

    release_region(advwdt_base_addr, advwdt_base_len);
    misc_deregister(&advwdt_miscdev);
    unregister_reboot_notifier(&advwdt_notifier);
}

module_init(advwdt_init);
module_exit(advwdt_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jianfeng.dai@advantech.com.cn");
MODULE_DESCRIPTION("Advantech WDT driver");



